@@ -15,7 +15,7 @@ from preseem_protobuf.model.element_pb2 import ElementDiscovery, ElementStatus
 # entities from being created.
 sync_lock = None
 
-# This object is used to pass element items to the sync_elem_discovery
+# This object is used to pass element items to the sync_preseem_model
 # algorithm for a given source and intent role.
 ElementSyncItem = namedtuple('ElementSyncItem', ('id', 'name', 'site', 'host'))
 
@@ -45,13 +45,14 @@ async def call_and_log_error(co):
     return None
 
 
-async def sync_elem_discovery(preseem_model,
-                              company_uuid,
-                              source,
-                              role,
-                              items,
-                              noact=False,
-                              verbose=0):
+async def sync_preseem_model(preseem_model,
+                             company_uuid,
+                             source,
+                             role,
+                             items,
+                             packages=None,
+                             noact=False,
+                             verbose=0):
     """
     Update the Preseem model to keep its ElementDiscovery objects in sync with
     another system.  This method takes the full set of items for a specific
@@ -71,10 +72,11 @@ async def sync_elem_discovery(preseem_model,
         raise ValueError('source cannot be the empty string')
 
     async with sync_lock:
-        model_discs, model_sites = await asyncio.gather(
+        model_discs, model_sites, model_packages = await asyncio.gather(
             preseem_model.Element.ListElementDiscovery(
                 company_uuid=company_uuid, intent_role_uuid=role.bytes),
-            preseem_model.Site.List(company_uuid=company_uuid))
+            preseem_model.Site.List(company_uuid=company_uuid),
+            preseem_model.Package.List(company_uuid=company_uuid))
 
         model_sites = {x.uuid: x for x in model_sites}
         model_discs = {x.uuid: x for x in model_discs}
@@ -110,7 +112,7 @@ async def sync_elem_discovery(preseem_model,
                     new_sites.add(item.site)
                 elif site.inactive:
                     logging.info(
-                        'sync_elem_discovery(%s): reactivate site "%s"',
+                        'sync_preseem_model(%s): reactivate site "%s"',
                         source, item.site)
                     if not noact:
                         await call_and_log_error(
@@ -119,12 +121,12 @@ async def sync_elem_discovery(preseem_model,
                                 uuid=site.uuid,
                                 inactive=False))
         if dup_ids:
-            logging.warning('sync_elem_discovery(%s): duplicate IDs found: %s',
+            logging.warning('sync_preseem_model(%s): duplicate IDs found: %s',
                             source, dup_ids)
             raise ValueError(f'duplicate IDs found')
 
         for site_name in new_sites:
-            logging.info('sync_elem_discovery(%s): create site "%s"', source,
+            logging.info('sync_preseem_model(%s): create site "%s"', source,
                          site_name)
             if not noact:
                 site = site_names[site_name] = await call_and_log_error(
@@ -161,7 +163,7 @@ async def sync_elem_discovery(preseem_model,
                     changes.append(f'site: {site.name} -> {item.site}')
                     mods['site_uuid'] = site_names[item.site].uuid
                 if mods:
-                    logging.info('sync_elem_discovery(%s): modify %s (%s)',
+                    logging.info('sync_preseem_model(%s): modify %s (%s)',
                                  source, item, ','.join(changes))
                     if not noact:
                         await call_and_log_error(
@@ -171,7 +173,7 @@ async def sync_elem_discovery(preseem_model,
                                 **mods))
 
             else:
-                logging.info('sync_elem_discovery(%s): create %s', source,
+                logging.info('sync_preseem_model(%s): create %s', source,
                              item)
                 if not noact:
                     site_uuid = site_names[item.site].uuid
@@ -187,13 +189,13 @@ async def sync_elem_discovery(preseem_model,
                     if disc:
                         uuid = UUID(bytes=disc.uuid)
                         logging.info(
-                            'sync_elem_discovery(%s): created ElementDiscovery "%s"',
+                            'sync_preseem_model(%s): created ElementDiscovery "%s"',
                             source, uuid)
                         uuid_map[item.id] = uuid
         for disc_uuid in unref_discs:
             disc = model_discs.get(disc_uuid)
             if not disc.inactive:
-                logging.info("sync_elem_discovery(%s): deactivate %s", source,
+                logging.info("sync_preseem_model(%s): deactivate %s", source,
                              UUID(bytes=disc_uuid))
                 if not noact:
                     await call_and_log_error(
@@ -201,4 +203,70 @@ async def sync_elem_discovery(preseem_model,
                             company_uuid=company_uuid,
                             uuid=disc_uuid,
                             inactive=True))
+
+        if packages is not None:
+            assert isinstance(packages, dict)
+            new_package_ids = set(packages)
+            for mpkg in model_packages:
+                if mpkg.source != source:
+                    continue
+                bpkg = packages.get(mpkg.source_id)
+                if bpkg:
+                    if not isinstance(bpkg.name, str) or bpkg.name == '':
+                        raise ValueError('invalid string value for package name')
+                    changes = []
+                    mods = {}
+                    new_package_ids.discard(mpkg.source_id)
+                    if mpkg.inactive:
+                        changes.append('reactivated')
+                        mods['inactive'] = False
+                    if mpkg.name != bpkg.name:
+                        changes.append(f'name: {mpkg.name} -> {bpkg.name}')
+                        mods['name'] = bpkg.name
+                    if mpkg.download_rate != bpkg.download_rate:
+                        changes.append(f'download_rate: {mpkg.download_rate} -> {bpkg.download_rate}')
+                        mods['download_rate'] = bpkg.download_rate
+                    if mpkg.upload_rate != bpkg.upload_rate:
+                        changes.append(f'upload_rate: {mpkg.upload_rate} -> {bpkg.upload_rate}')
+                        mods['upload_rate'] = bpkg.upload_rate
+                    if mods:
+                        logging.info('sync_packages(%s): modify %s (%s)',
+                                     source, mpkg, ', '.join(changes))
+                        await preseem_model.Package.Update(
+                            company_uuid=company_uuid,
+                            uuid=mpkg.uuid,
+                            **mods)
+                else:
+                    if not mpkg.inactive:
+                        logging.info("sync_preseem_model(%s): deactivate package %s", source,
+                                     UUID(bytes=mpkg.uuid))
+                        if not noact:
+                            await call_and_log_error(
+                                preseem_model.Package.Update(
+                                    company_uuid=company_uuid,
+                                    uuid=mpkg.uuid,
+                                    inactive=True))
+            for pkg_id in new_package_ids:
+                bpkg = packages.get(pkg_id)
+                if bpkg:
+                    if not isinstance(bpkg.id, str) or bpkg.id == '':
+                        raise ValueError('invalid string value for package source_id')
+                    if not isinstance(bpkg.name, str) or bpkg.name == '':
+                        raise ValueError('invalid string value for package name')
+                    logging.info('sync_preseem_model(%s): create %s', source, bpkg)
+                    if not noact:
+                        mpkg = await call_and_log_error(
+                            preseem_model.Package.Create(
+                                company_uuid=company_uuid,
+                                name=bpkg.name,
+                                source=source,
+                                source_id=bpkg.id,
+                                download_rate=bpkg.download_rate,
+                                upload_rate=bpkg.upload_rate))
+                        if mpkg:
+                            uuid = UUID(bytes=mpkg.uuid)
+                            logging.info(
+                                'sync_preseem_model(%s): created Package "%s"',
+                                source, uuid)
+
         return uuid_map
